/*  Gtk+ Widget Factory
 *  Copyright (C) 1999-2000  Jeroen Benckhuijsen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "gwf_projectio.h"
#include "gwf_error.h"
#include "gwf_project.h"
#include "gwf_source.h"
#include "gwf_format.h"
#include "gwf_projlist.h"
#include "gwf_license.h"
#include "gwf_xmlio.h"
#include <gtk/gtk.h>
#include <string.h>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

static gint	backup_file			(gchar *ifname);
static gint	check_srcfile_backup		(GwfProject *project);
static gchar*	get_full_path_source_file	(GwfProject *project,gchar *sourcefile);

gint
gwf_save_project(GwfError *error,GwfProjlist *projlist,GwfXmlIO *xmlio)
{
  GwfProject	*project;
  gchar 	*filename;

  g_return_val_if_fail(error != NULL,0);
  g_return_val_if_fail(projlist != NULL,0);
  g_return_val_if_fail(xmlio != NULL,0);

  project = gwf_projlist_get_current(projlist);
  if (!project)
  {
    gwf_error_raise_error(error,"General : No project");
    return 0;
  }

  if (!gwf_xmlio_save(xmlio,project))
  {
    gwf_error_raise_error(error,"XmlIO : Save : Conversion error");
    return 0;
  }

  filename = gwf_project_get_filename(project);
  if (!filename)
  {
    gwf_error_raise_error(error,"General : Save : No filename");
    return 0;
  }

  if (!gwf_format_save(error,filename,gwf_xmlio_get_buffer(xmlio)))
    return 0;

  gwf_project_set_changed(project,FALSE);
  gwf_error_raise_error(error,"General : Project saved");

  return 1;
}

void
gwf_load_project (GwfError *error,GwfProjlist *projlist,GwfXmlIO *xmlio,gchar *filename)
{
  gchar			*xmlbuf = NULL;
  GwfProject		*project;

  g_return_if_fail(error != NULL);
  g_return_if_fail(projlist != NULL);
  g_return_if_fail(xmlio != NULL);
  g_return_if_fail(filename != NULL);

  if (!gwf_format_load(error,filename,&xmlbuf))
    return;

  if (!xmlbuf)
  {
    gwf_error_raise_error(error,"General : Load : no XML");
    return;
  }

  gwf_xmlio_set_buffer(xmlio,xmlbuf);
  project = gwf_xmlio_load(xmlio,projlist,error);
  if (!project)
  {
    gwf_error_raise_error(error,"General : Load : Invalid XML");
    return;
  }

  gwf_project_set_filename(project,filename);
  gwf_project_set_changed(project,FALSE);

  gwf_error_raise_error(error,"General : Project loaded");
}

void
gwf_write_project_source	(GwfError *error,GwfProjlist *projlist,GwfSource *source,GwfLicList *liclist)
{
  GwfProject		*project;
  FILE			*hfile,*cfile;
  gchar			*buffer;
  gchar			*fname;

  g_return_if_fail(error != NULL);
  g_return_if_fail(projlist != NULL);
  g_return_if_fail(source != NULL);
  g_return_if_fail(liclist != NULL);

  project = gwf_projlist_get_current(projlist);
  if (!project)
  {
    gwf_error_raise_error(error,"General : No project");
    return;
  }

  if (!gwf_project_get_cfile(project))
  {
    gwf_error_raise_error(error,"General : Source write : No C-file");
    return;
  }

  if (!gwf_project_get_hfile(project))
  {
    gwf_error_raise_error(error,"General : Source write : No H-file");
    return;
  }

  if (!gwf_source_write_check(source,project,error))
  {
    gwf_error_raise_error(error,"General : Source write : Check error");
    return;
  }

  if (!check_srcfile_backup(project))
  {
    gwf_error_raise_error(error,"General : Source write : Backup error");
    return;
  }

  fname = get_full_path_source_file(project,gwf_project_get_hfile(project));
  hfile = fopen(fname,"w");
  g_free(fname);

  if (!hfile)
  {
    gwf_error_raise_error(error,"General : Source write : Can't open H-file");
    return;
  }

  fname = get_full_path_source_file(project,gwf_project_get_cfile(project));
  cfile = fopen(fname,"w");
  g_free(fname);

  if (!cfile)
  {
    gwf_error_raise_error(error,"General : Source write : Can't open C-file");
    return;
  }

  buffer = gwf_source_write_h(source,project,liclist);
  if (fwrite((void*)buffer,sizeof(gchar),strlen(buffer),hfile) != strlen(buffer))
  {
    gwf_error_raise_error(error,"General : Source write : Can't write H-file");
    g_free(buffer);
    return;
  }
  g_free(buffer);

  buffer = gwf_source_write_c(source,project,liclist);
  if (fwrite((void*)buffer,sizeof(gchar),strlen(buffer),cfile) != strlen(buffer))
  {
    gwf_error_raise_error(error,"General : Source write : Can't write C-file");
    g_free(buffer);
    return;
  }
  g_free(buffer);

  fclose(hfile);
  fclose(cfile);

  gwf_error_raise_error(error,"General : Source written");
}

static gint
check_srcfile_backup(GwfProject *project)
{
  gchar *fname;

  if (!gwf_project_get_backup(project))
    return 1;

  fname = get_full_path_source_file(project,gwf_project_get_cfile(project));
  if (!backup_file(fname))
  {
    g_free(fname);
    return 0;
  }
  g_free(fname);

  fname = get_full_path_source_file(project,gwf_project_get_hfile(project));
  if (!backup_file(fname))
  {
    g_free(fname);
    return 0;
  }
  g_free(fname);

  return 1;
}

static gint
backup_file(gchar *ifname)
{
  FILE 		*ifp,*ofp;
  gchar		*ofname,c;


  ifp = fopen(ifname,"r");
  if (ifp)
  {
    ofname = g_strdup_printf("%s.bak",ifname);
    ofp = fopen(ofname,"w");
    if (!ofp)
      return 0;

    while ((c = getc(ifp)) != EOF)
      putc(c,ofp);

    fclose(ifp);
    fclose(ofp);
    g_free(ofname);
  }
  return 1;
}

void
gwf_project_filname_check_extension_add(GwfProject *project)
{
  gchar		*tmp;
  gchar		*source;

  g_return_if_fail(project != NULL);

  source = g_strdup(gwf_project_get_filename(project));
  if (!source)
    return;

  tmp = source + strlen(source) - strlen(".gwf");
  if ((tmp > source) && (strcmp(tmp,".gwf") == 0))
  {
    g_free(source);
    return;
  }

  tmp = g_strdup_printf("%s/%s",source,".gwf");
  gwf_project_set_filename(project,tmp);
  g_free(source);
}

static gchar*
get_full_path_source_file(GwfProject *project,gchar *sourcefile)
{
  gchar		*base;
  gchar		*ret;

  g_return_val_if_fail(project != NULL,NULL);

  if (!gwf_project_get_filename(project))
    return g_strdup(sourcefile);

  if ((*sourcefile == G_DIR_SEPARATOR) || (*sourcefile == '.'))
    return g_strdup(sourcefile);

  base = g_dirname(gwf_project_get_filename(project));
  ret = g_strdup_printf("%s%c%s",base,G_DIR_SEPARATOR,sourcefile);
  g_free(base);

  return ret;
}
