/*  Gtk+ Widget Factory
 *  Copyright (C) 1999-2000  Jeroen Benckhuijsen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <gtk/gtk.h>
#include "gwf_treedata.h"
#include "gwf_project.h"
#include "gwf_tools.h"

typedef struct _GwfTreeDataForeach	GwfTreeDataForeach;

struct _GwfTreeDataForeach
{
  GwfTreeDataForeachFunc	func;
  gpointer			user_data;
};

static void		foreach_treedata 		(gpointer key,gpointer value,gpointer user_data);
static gboolean		tree_data_free 			(gpointer key,gpointer value,gpointer user_data);

static gint		gwf_treedata_connect_ptdata 	(GwfTreeData *data,GwfProject *project,GwfPTData *ptdata);
static void		gwf_treedata_disconnect_ptdata	(GwfTreeData *data,gint node,GwfProject *project);
static GwfPTData*	gwf_treedata_get_ptdata		(GwfTreeData *data,gint node);
static void		project_destroyed_cb		(GtkObject *object,gpointer user_data);
static void		remove_ptdata_foreach_node	(GtkCTree *tree,GtkCTreeNode *node,gpointer user_data);

GwfTreeDataList*
gwf_treedata_list_new (void)
{
  GwfTreeDataList	*tmp;
  
  tmp = g_malloc(sizeof(GwfTreeDataList));
  
  tmp->datas = g_hash_table_new(g_str_hash,g_str_equal);
  return tmp;
}

void
gwf_treedata_list_free	(GwfTreeDataList *list)
{
  g_return_if_fail(list != NULL);
  
  g_hash_table_foreach_remove(list->datas,tree_data_free,NULL);
  g_hash_table_destroy(list->datas);
  g_free(list);
}

gint
gwf_treedata_list_add (GwfTreeDataList *list,GwfTreeData *data)
{
  g_return_val_if_fail(list != NULL,0);
  g_return_val_if_fail(data != NULL,0);
  
  if (g_hash_table_lookup(list->datas,(gpointer)data->name))
    return 0;
  
  g_hash_table_insert(list->datas,(gpointer)data->name,(gpointer)data);
  return 1;
}

void
gwf_treedata_list_remove (GwfTreeDataList *list,gchar *name)
{
  GwfTreeData	*tmp;
  g_return_if_fail(list != NULL);
  g_return_if_fail(name != NULL);
  
  tmp = (GwfTreeData*)g_hash_table_lookup(list->datas,(gconstpointer)name);
  if (!tmp)
    return;
   
  g_hash_table_remove(list->datas,(gconstpointer)name);
  gwf_treedata_free(tmp);
}

GwfTreeData*
gwf_treedata_list_get_data (GwfTreeDataList *list,gchar *name)
{
  g_return_val_if_fail(list != NULL,NULL);
  g_return_val_if_fail(name != NULL,NULL);
  
  return (GwfTreeData*)g_hash_table_lookup(list->datas,(gconstpointer)name);
}

void
gwf_treedata_list_foreach (GwfTreeDataList *list,GwfTreeDataForeachFunc func,gpointer user_data)
{
  GwfTreeDataForeach	tmp;
  
  g_return_if_fail(list != NULL);
  g_return_if_fail(func != NULL);  
  
  tmp.func = func;
  tmp.user_data = user_data;
  
  g_hash_table_foreach(list->datas,foreach_treedata,(gpointer)&tmp);
}

GwfTreeData*
gwf_treedata_new (gchar *name)
{
  GwfTreeData		*tmp;
  
  g_return_val_if_fail(name != NULL,NULL);
  
  tmp = g_malloc(sizeof(GwfTreeData));
  
  tmp->name = g_strdup(name);
  tmp->tree = NULL;
  tmp->attach = NULL;
  tmp->free = NULL;
  tmp->ptdata = g_ptr_array_new();
  tmp->nnodes = 0;
  
  return tmp;
}

void
gwf_treedata_free (GwfTreeData *data)
{
  g_return_if_fail(data != NULL);

  g_ptr_array_free(data->ptdata,FALSE);
  g_free(data->name);
  g_free(data);
}

void
gwf_treedata_set_attach	(GwfTreeData *data,GwfTreeDataAttachFunc func)
{
  g_return_if_fail(data != NULL);
  
  data->attach = func;
}

void
gwf_treedata_set_free (GwfTreeData *data,GwfTreeDataFreeFunc func)
{
  g_return_if_fail(data != NULL);

  data->free = func;
}

void
gwf_treedata_attach (GwfTreeData *data,GwfProject *project,GtkCTree *tree)
{
  g_return_if_fail(data != NULL);
  g_return_if_fail(project != NULL);
  g_return_if_fail(tree != NULL);

  data->tree = tree;
  data->attach(data,project);
}

void
gwf_treedata_free_data(GwfTreeData *data)
{
  g_return_if_fail(data != NULL);

  if (data->free)
    data->free(data);
}

static void
foreach_treedata (gpointer key,gpointer value,gpointer user_data)
{
  GwfTreeDataForeach	*tmp = user_data;

  tmp->func((GwfTreeData*)value,tmp->user_data);
}

static gboolean
tree_data_free (gpointer key,gpointer value,gpointer user_data)
{
  gwf_treedata_free((GwfTreeData*)value);
  return TRUE;
}









void
gwf_treedata_unregister_all_ptdata(GwfTreeData *data)
{
  GwfPTData	*ptdata;
  gint		i;

  for (i = 0 ; i < data->nnodes ; i++)
  {
    ptdata = gwf_treedata_get_ptdata(data,i);
    if (ptdata)
      gwf_treedata_disconnect_ptdata(data,i,ptdata->project);
    else
      g_warning("ptdata not found!!!!!");
  }
  gwf_treedata_free_data(data);
}

void
gwf_treedata_unregister_ptdata_by_node(GwfTreeData *data,GtkCTreeNode *node)
{
  gint		i;
  GwfPTData	*ptdata;

  for (i = data->nnodes-1 ; i >=0 ; i--)
  {
    ptdata = gwf_treedata_get_ptdata(data,i);
    if (ptdata)
    {
      if (ptdata->parent == node)
        gwf_treedata_disconnect_ptdata(data,i,ptdata->project);
    }
    else
      g_warning("ptdata not found!!!!!");
  }
}

GwfPTData**
gwf_treedata_register_ptdata(GwfTreeData *data,GwfPTData def[],gint nnodes,GtkCTreeNode *parent,GwfProject *project)
{
  GwfPTData	**pt_data;
  gint		i;

  g_return_val_if_fail(data != NULL,NULL);
  g_return_val_if_fail(def != NULL,NULL);
  g_return_val_if_fail(nnodes > 0,NULL);
  g_return_val_if_fail(parent != NULL,NULL);
  g_return_val_if_fail(project != NULL,NULL);

  pt_data = g_malloc(sizeof(GwfPTData*) * nnodes);
  for (i = 0 ; i < nnodes ; i++)
  {
    pt_data[i] = g_malloc(sizeof(GwfPTData));
    pt_data[i]->data = def[i].data;
    pt_data[i]->node = def[i].node;
    pt_data[i]->parent = def[i].parent;
    pt_data[i]->signame = def[i].signame;
    pt_data[i]->sigid = def[i].sigid;
    pt_data[i]->func = def[i].func;
    pt_data[i]->user_data = def[i].user_data;
    pt_data[i]->project = project;
    pt_data[i]->data = data;
    pt_data[i]->parent = parent;
  }

  data->nnodes += nnodes;
  for ( i = 0 ; i < nnodes ; i++ )
    g_ptr_array_add(data->ptdata,(gpointer)pt_data[i]);

  gtk_signal_connect(GTK_OBJECT(project),"destroy",(GtkSignalFunc)project_destroyed_cb,(gpointer)data);

  for (i = 0 ; i < nnodes ; i++)
    g_return_val_if_fail(gwf_treedata_connect_ptdata(data,project,pt_data[i]),NULL);

  return pt_data;
}

static GwfPTData*
gwf_treedata_get_ptdata	(GwfTreeData *data,gint node)
{
  g_return_val_if_fail(data != NULL,NULL);
  g_return_val_if_fail((node >= 0) || (node < data->nnodes), NULL);

  return (GwfPTData*)g_ptr_array_index(data->ptdata, node);
}

static gint
gwf_treedata_connect_ptdata (GwfTreeData *data,GwfProject *project,GwfPTData *ptdata)
{
  gchar		*signame;

  g_return_val_if_fail(data != NULL,0);
  g_return_val_if_fail(project != NULL,0);
  g_return_val_if_fail(ptdata != NULL,0);

  signame = (ptdata->signame) ? ptdata->signame : gtk_signal_name(ptdata->sigid);
  g_return_val_if_fail(signame != NULL,0);

  gtk_signal_connect(GTK_OBJECT(project),signame,ptdata->func,(gpointer)ptdata);
  return 1;
}

static void
gwf_treedata_disconnect_ptdata (GwfTreeData *data,gint node,GwfProject *project)
{
  g_return_if_fail(data != NULL);
  g_return_if_fail((node >= 0) || (node < data->nnodes));

  gtk_signal_disconnect_by_data(GTK_OBJECT(project),(gpointer)gwf_treedata_get_ptdata(data,node));
  g_ptr_array_remove_index(data->ptdata,node);
  data->nnodes--;
}


static void
project_destroyed_cb(GtkObject *object,gpointer user_data)
{
  GwfTreeData	*data = user_data;
  GwfProject	*project = (GwfProject*)object;
  GwfPTData	*ptdata;
  gint		i;

  for (i = data->nnodes-1 ; i >= 0 ; i--)
  {
    ptdata = gwf_treedata_get_ptdata(data,i);
    if (ptdata)
    {
      if (ptdata->project == project)
      {
        gwf_treedata_disconnect_ptdata(data,i,(GwfProject*)object);
      }
    }
    else
      g_warning("ptdata not found!!!!!");
  }
}

/**********************************************/
/* Standard node insert function by GwfPTData */
/**********************************************/

void
gwf_projtree_remove_node_ptdata(GwfTreeData *data,GtkCTreeNode *node)
{
  g_return_if_fail(data != NULL);
  g_return_if_fail(node != NULL);

  gtk_ctree_pre_recursive(data->tree,node,remove_ptdata_foreach_node,(gpointer)data);

  gtk_ctree_remove_node(data->tree,node);
}

static void
remove_ptdata_foreach_node(GtkCTree *tree,GtkCTreeNode *node,gpointer user_data)
{
  gwf_treedata_unregister_ptdata_by_node((GwfTreeData*)user_data,node);
}

GwfPTData**
gwf_projtree_add_new_section (GwfTreeData *data,GwfProject *project,gchar *secname,GwfPTData def[],gint nnodes)
{
  GtkCTreeNode	*parent;
  GtkCTreeNode	*section;

  parent = gtk_ctree_find_project(data->tree,project);
  if (!parent)
    return NULL;

  section = gtk_ctree_add_node(data->tree,parent,secname);
  if (!section)
    return NULL;

  return gwf_treedata_register_ptdata(data,def,nnodes,section,project);
}

void
gwf_projtree_insert_string_by_ptdata(GwfPTData *ptdata,gchar *value)
{
  g_return_if_fail(ptdata != NULL);
  g_return_if_fail(ptdata->data != NULL);
  g_return_if_fail(ptdata->parent != NULL);

  gtk_ctree_insert_text_value(ptdata->data->tree,ptdata->parent,ptdata->node,value);
}

void
gwf_projtree_insert_bool_by_ptdata(GwfPTData *ptdata,gboolean value)
{
  g_return_if_fail(ptdata != NULL);
  g_return_if_fail(ptdata->data != NULL);
  g_return_if_fail(ptdata->parent != NULL);

  gtk_ctree_insert_bool_value(ptdata->data->tree,ptdata->parent,ptdata->node,value);
}

void
gwf_projtree_insert_float_by_ptdata(GwfPTData *ptdata,gfloat value)
{
  g_return_if_fail(ptdata != NULL);
  g_return_if_fail(ptdata->data != NULL);
  g_return_if_fail(ptdata->parent != NULL);

  gtk_ctree_insert_float_value(ptdata->data->tree,ptdata->parent,ptdata->node,value);
}

void
gwf_projtree_insert_int_by_ptdata(GwfPTData *ptdata,glong value)
{
  g_return_if_fail(ptdata != NULL);
  g_return_if_fail(ptdata->data != NULL);
  g_return_if_fail(ptdata->parent != NULL);

  gtk_ctree_insert_int_value(ptdata->data->tree,ptdata->parent,ptdata->node,value);
}

/**********************************/
/* Standard node change callbacks */
/**********************************/

void
gwf_projtree_node_string_changed_cb (GwfProject *project,gchar *text,gpointer user_data)
{
  GwfPTData	*pt_data = user_data;
  GtkCTreeNode	*node;

  node = gtk_ctree_find_item_part(pt_data->data->tree,pt_data->parent,pt_data->node);
  g_return_if_fail(node != NULL);

  gtk_ctree_alter_text_value(pt_data->data->tree,node,pt_data->node,text);
}

void
gwf_projtree_node_bool_changed_cb (GwfProject *project,gboolean bool,gpointer user_data)
{
  GwfPTData	*pt_data = user_data;
  GtkCTreeNode	*node;

  node = gtk_ctree_find_item_part(pt_data->data->tree,pt_data->parent,pt_data->node);
  g_return_if_fail(node != NULL);

  gtk_ctree_alter_bool_value(pt_data->data->tree,node,pt_data->node,bool);
}

void
gwf_projtree_node_int_changed_cb (GwfProject *project,glong value,gpointer user_data)
{
  GwfPTData	*pt_data = user_data;
  GtkCTreeNode	*node;

  node = gtk_ctree_find_item_part(pt_data->data->tree,pt_data->parent,pt_data->node);
  g_return_if_fail(node != NULL);

  gtk_ctree_alter_int_value(pt_data->data->tree,node,pt_data->node,value);
}

void
gwf_projtree_node_float_changed_cb (GwfProject *project,gfloat value,gpointer user_data)
{
  GwfPTData	*pt_data = user_data;
  GtkCTreeNode	*node;

  node = gtk_ctree_find_item_part(pt_data->data->tree,pt_data->parent,pt_data->node);
  g_return_if_fail(node != NULL);

  gtk_ctree_alter_float_value(pt_data->data->tree,node,pt_data->node,value);
}
